> module Evaluate
> where

> import Prelude hiding (lookup)
> import qualified Prelude
> import Data.Char
> import Data.Maybe
> import Data.List (sort, nub)
> import System.IO
> import Control.Monad (when, mplus, msum, liftM2)
> import Data.Array
> import Data.List (genericLength)
> import Data.Set (member, union, difference, singleton, unions, Set)
> import Data.IORef
> import qualified System.Console.Readline as Readline
> import System

> import Control.Monad.Reader hiding (join)
> import Printer (keyword', keyword, ppCon, ppCol, P(..), withPrec, parensPrec, defaultPState, PState(..)) 
> import Message hiding (empty)

> import Options hiding (debug)
> import qualified Options as O
> import Core
> -- import qualified Debug.Trace

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Values}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data Val
>   =  VBool   !Bool
>   |  VNat    !Integer
>   |  VString !String
>   |  VType -- type argument marker
>   |  Closure Pat Expr Bool !Env -- recursive environment? 
>   |  VRecord ![Row Val]
>   |  VCon    !Con !Val
>   |  Addr    !Integer
>   |  Cont    !Stack
>   |  VEnv    !Env -- first-class environments
>   |  VClass  Expr Env
>   |  VObject !Env -- objects
>   |  VSuper  !Env -- objects
>   |  VMethod Expr !Env
>   |  VArray  !(Array Integer Val)
>   |  VHole   (IORef (Maybe Val))

To create a sorted/normalized record value after evaluation of the
components:

> vrecord :: [Row Val] -> Val
> vrecord rs = VRecord (sort rs)

NB. `sort' must be a stable sorting algorithm, so that the relative
order of numbered fields is not changed.

Die Konstruktoren sind strikt in allen Argumenten (damit
z.B. arithmetische Operationen auch wirklich ausgefhrt werden).

> pVal :: Options -> Int -> Val -> Doc
> pVal o p v = runReader (ppVal v) (defaultPState { pcolor = O.color o, pdebug = O.debug o, pprec = p })

> ppVal :: Val -> P Doc
> ppVal (VBool b) 
>   =  ppBool b
> ppVal (VNat n) 
>   =  return (pretty n)
> ppVal (VString s) 
>   =  return (pretty s)
> ppVal VType
>   =  do
>        pc <- ppCol astype
>        return (pc (text "<>"))
> ppVal (Closure p e _ env)
>   =  parensPrec 10 $ do
>        pp <- withPrec 0 $ ppPat p
>        pe <- withPrec 0 $ ppExpr e
>        penv <- ppEnv_ (difference (free e) (freePat p)) env
>        pa <- alignkey "fun" (pp <> text " =>" <+> pe)
>        return (group (pa <+> penv)) -- subtle because it could loop for recursive closures, but it doesn't right now due to VHole
> ppVal (VRecord rs)
>   =  do
>        prs <- withPrec 0 $ mapM (ppRow ppVal) rs
>        return (parens (group (intersperse (char ',' <> sp) prs)))

> ppVal (VCon c (VRecord [])) 
>   =  ppCon c
> ppVal (VCon c v) 
>   =  parensPrec 10 $ do
>        pc <- ppCon c
>        pv <- withPrec 11 $ ppVal v
>        return (group (pc <+> pv))
> ppVal (Addr a) 
>   =  return (text "&" <> pretty a)
> ppVal (Cont ctl)
>   =  ppStack ctl
> ppVal (VEnv env)
>   =  ppEnv env
> ppVal (VClass e env)
>   =  do
>        pk <- keyword' "class"
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv' env
>        return (group (pk <+> pe <+> penv))
> ppVal (VObject env)
>   =  do
>        pk <- keyword' "object"
>        penv <- ppEnv' env
>        return (group (pk <+> penv))
> ppVal (VSuper env)
>   =  do
>        pk <- keyword' "super"
>        penv <- ppEnv' env
>        return (group (pk <+> penv))
> ppVal (VMethod e env)
>   =  do
>        pk <- keyword' "method"
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv' env
>        return (group (pk <+> pe <+> penv))
> ppVal (VArray a)
>   =  ppList ppVal (elems a)

 ppVal e -- catch-all (vorlufig)
   =  pretty e

> ppVal (VHole _)
>   =  keyword' "rec"

> instance Show Val where
>   showsPrec _ (VBool b) = shows b
>   showsPrec _ (VNat n) = shows n
>   showsPrec _ (VString s) = shows s
>   showsPrec _ (Closure x e _ _)
>     = showString "<" . shows x . showString ", " . shows e . showString ", [..]>"
>   showsPrec _ (VRecord fs)
>     =  shows fs
>   showsPrec _ (VCon c v) 
>     = showString "(" . showString c .showString " " . shows v . showString ")"
>   showsPrec _ (Cont ctl)
>     = showString "<" . shows ctl . showString ">"

> instance Pretty Val

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Environment}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> newtype Env
>   =  Env [(Var, Val)]
>      deriving (Show)

> lookup x (Env env) 
>   =  Prelude.lookup x env

> empty = Env []
> single b = Env [b]

> Env env1 \+/ Env env2 = Env (env1 ++ env2) -- inverse (!) comma operator

> --reverseCat :: [a] -> [a] -> [a]
> reverseCat [] y = y
> reverseCat (a : x) (Env y) = reverseCat x (Env (a : y))

> toList (Env env) = env

> domain (Env env) = map fst env

> ppEnv', ppEnv :: Env -> P Doc
> ppEnv' _
>   =  return (text "[..]") -- ppList ppBind env
> ppEnv (Env env)
>   =  ppList ppBind env

Print an environment restricted to a set of identifiers:

> ppEnv_ :: Set Var -> Env -> P Doc
> ppEnv_ v (Env env)
>   =  ppList ppBind (filter (\ (x, _) -> member x v) env)

> ppBind :: (Var, Val) -> P Doc
> ppBind (x, v)
>   =  fmap (align (x ++ " |-> ")) (ppVal v)

 align s d = block (length s) (text s <> d)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Stack frames}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type Stack
>   =  [Frame]

> data Frame
>   =  If1 {--} Expr Expr Env
>   |  UnOp1 UnOp {--}
>   |  BinOp1 {--} BinOp Expr Env
>   |  BinOp2 Val BinOp {--}
>   |  Let1 {--} Expr Env
>   |  Val1 Pat {--}
>   |  App1 {--} Expr Env
>   |  App2 Val  {--}
>   |  TApp2 {--} Val -- eine Variante von `App2': beim Einfuegen eines Typarguments ist das Funktionsargument anschliessend schon ausgewertet
>   |  Rec1 Pat {--} Env
>   |  Recordi [Row Val] (Row ()) {--} [Row Expr] Env
>   |  Recordn [Row Val] (Row ()) {--}
>   |  Sel1 {--} Lab Env -- `Env' wird fr die Implementierung von `super' bentigt
>   |  Con1 Con {--}
>   |  Seq1 {--} Expr Env
>   |  Ref1 {--}
>   |  Assign1 {--} Expr Env
>   |  Assign2 Val {--}
>   |  Deref1 {--}
>   |  Case1 {--} [(Pat, Expr)] Env
>   |  Continue1 {--} Expr Env
>   |  Continue2 Val {--}
>   |  Throw1 {--}
>   |  Try1 {--} [(Pat, Expr)] Env
>   |  Call1 Prim {--}
>   |  While1 Expr Expr Env

>   |  Comma1 {--} Expr Env
>   |  Comma2 Val {--}
>   |  Lookup1 {--} Var

>   |  Object1 {--}
>   |  New1 {--}
>   |  Invoke1 {--} Lab Expr Env
>   |  Invoke2 Val Lab {--}
>   |  Inherit1 {--}
>   |  InheritIn1 Var {--} Expr Env
>   |  InheritIn2 Var {--} Expr Env
>   |  InheritIn3 Var Val {--}
>   |  Include1 {--}
>   |  IncludeIn1 Var {--} Expr Env
>   |  IncludeIn2 Var Val {--}
>   |  Append1 {--} Expr Env
>   |  Append2 Val {--}

>   |  ArrFun1 {--} Var Expr Env
>   |  ArrFun2i !Integer !Integer [Val] Var Expr Env
>   |  ArrListi [Val] {--} [Expr] Env
>   |  ArrListn [Val] {--}
>   |  Subscript1 {--} Expr Env
>   |  Subscript2 Val {--}
>      deriving (Show)

> instance Pretty Frame

> hole :: Doc
> hole = text "(@)"

> ppFrame :: Frame -> P Doc
> ppFrame (If1 {--} e2 e3 env)
>   =  do
>        pki <- keyword "if"
>        pkt <- keyword "then"
>        pke <- keyword "else"
>        pe2 <- withPrec 0 $ ppExpr e2
>        pe3 <- withPrec 0 $ ppExpr e3
>        penv <- ppEnv_ (union (free e2) (free e3)) env
>        return (group (pki <> 
>                       block 3 (hole <+> block 5 (pkt <> pe2) <+>
>                                         block 5 (pke <> pe3)) <+> penv))
> ppFrame (UnOp1 op {--})
>   =  return (group (pretty op <+> hole))
> ppFrame (BinOp1 {--} op e env)
>   =  parensPrec 10 $ do
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv_ (free e) env
>        return (group (hole <> text " " <> pretty op <+> block 2 pe <+> penv))
> ppFrame (BinOp2 v op {--})
>   =  parensPrec 10 $ do
>        pv <- withPrec 11 $ ppVal v
>        return (group (pv <> text " " <> pretty op <+> block 2 hole))
> ppFrame (Let1 {--} e env)
>   =  do
>        pkl <- keyword "let"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pe <- withPrec 0 $ ppExpr e
>        -- TODO: ppEnv_ (difference (free e) (freePat p)) env))
>        return (group (pkl <> hole <> 
>                       block 1 (sp <> pki <> block 3 pe) <+> pke)) -- TODO: <+> penv
> ppFrame (Val1 p {--})
>   =  do
>        pk <- keyword "val"
>        pp <- withPrec 0 $ ppPat p
>        return (group (pk <> block 4 (pp <+> text "=" <+> hole)))
> ppFrame (App1 {--} e env)
>   =  parensPrec 10 $ do
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv_ (free e) env
>        return (group (hole <> text " @" <+> pe <+> penv))
> ppFrame (App2 v {--})
>   =  parensPrec 10 $ do
>        pv <- withPrec 10 $ ppVal v
>        return (group (pv <> text " @" <+> hole))
> ppFrame (TApp2 {--} v)
>   =  parensPrec 10 $ do
>        pv <- withPrec 11 $ ppVal v
>        return (group (hole <> text " @" <+> pv))
> ppFrame (Rec1 p {--} env)
>   =  parensPrec 10 $ do
>        pk <- keyword "rec"
>        pp <- withPrec 0 $ ppPat p
>        penv <- ppEnv env
>        return (group (pk <> pp <> text " => " <> hole <+> penv))
> ppFrame (Recordi rs1 r rs2 env)
>   =  do
>        prs1 <- withPrec 0 $ mapM (ppRow ppVal) rs1
>        prsh <- mapM (ppRow (const (return hole))) [r]
>        prs2 <- withPrec 0 $ mapM (ppRow ppExpr) rs2
>        penv <- ppEnv_ (unions (map (freeRow free) rs2)) env
>        return (group (parens (group (intersperse (char ',' <> sp) 
>                      (prs1 ++ prsh ++ prs2))) <+> 
>                      penv))
> ppFrame (Recordn rs1 r)
>   =  do
>        prs1 <- mapM (ppRow ppVal) rs1
>        prs2 <- mapM (ppRow (const (return hole))) [r]
>        return (parens (group (intersperse (char ',' <> sp) 
>                              (prs1 ++ prs2))))
> ppFrame (Sel1 {--} l _)
>   =  return (group (hole <> text "." <> ppLab l)) -- what about env?
> ppFrame (Con1 c {--})
>   =  parensPrec 10 $ do
>        pc <- ppCon c
>        return (group (pc <+> hole))
> ppFrame (Seq1 {--} e env)
>   =  parensPrec 10 $ do
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv_ (free e) env
>        return (group (hole <> text ";" <+> pe <+> penv))
> ppFrame (Ref1 {--})
>   =  do
>        pk <- keyword' "ref"
>        parensPrec 10 $
>          return (group (pk <+> hole))
> ppFrame (Assign1 {--} e env)
>   =  parensPrec 10 $ do
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv_ (free e) env
>        return (group (hole <+> text ":=" <+> pe <+> penv))
> ppFrame (Assign2 v {--})
>   =  parensPrec 10 $ do
>        pv <- withPrec 11 $ ppVal v
>        return (group (pv <+> text ":=" <+> hole))
> ppFrame (Deref1 {--})
>   =  parensPrec 10 $
>        return (group (text "!" <> hole))
> ppFrame (Case1 {--} rs env)
>   =  do
>        pkc <- keyword "case"
>        pko <- keyword "of"
>        pke <- keyword' "end"
>        prs <- withPrec 0 $ mapM ppRule rs
>        penv <- ppEnv env
>        return (group (pkc <> hole <+> pko <>
>                       intersperse (sp <> text " | ") prs <+> pke <+> penv))
> ppFrame (Continue1 {--} e env)
>   =  do
>        pk <- keyword' "continue"
>        parensPrec 10 $ do
>          pe <- withPrec 11 $ ppExpr e
>          penv <- ppEnv_ (free e) env
>          return (group (pk <+> hole <+> pe <+> penv))
> ppFrame (Continue2 v {--})
>   =  do
>        pk <- keyword' "continue"
>        parensPrec 10 $ do
>          pv <- ppVal v
>          return (group (pk <+> pv <+> hole))
> ppFrame (Try1 {--} rs env)
>   =  do
>        pkt <- keyword' "try"
>        pkc <- keyword' "catch"
>        pke <- keyword' "end"
>        prs <- withPrec 0 $ mapM ppRule rs
>        penv <- ppEnv env
>        return (group (pkt <> block 2 (sp <> hole) <+> pkc <+>
>                              block 2 (intersperse (sp <> text " | ") prs) <+> pke <+> penv))
> ppFrame (Throw1 {--})
>   =  do
>        pk <- keyword' "throw"
>        parensPrec 10 $
>          return (group (pk <+> hole))
> ppFrame (Call1 f {--})
>   =  parensPrec 10 $
>        return (group (pretty f <+> hole))
> --ppFrame (Show1 {--})
> --  =  group (text "show" <+> hole)
> ppFrame (While1 e1 e2 env)
>   =  do
>        pkw <- keyword "while"
>        pkd <- keyword' "do"
>        pke <- keyword' "end"
>        pe1 <- withPrec 0 $ ppExpr e1
>        pe2 <- withPrec 0 $ ppExpr e2
>        penv <- ppEnv_ (union (free e1) (free e2)) env
>        return (group (pkw <> 
>                       block 2 (pe1 <> sp <> pkd <+> pe2) <+> penv))

> ppFrame (Comma1 {--} e env)
>   =  do
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv_ (free e) env
>        return (group (hole <+> pe <+> penv))
> ppFrame (Comma2 v {--})
>   =  do
>        pv <- withPrec 11 $ ppVal v
>        return (group (pv <+> hole))
> ppFrame (Lookup1 {--} x)
>   =  return (group (hole <> text "$" <> text x))

> ppFrame (Object1 {--})
>   =  do
>        pko <- keyword "object"
>        pke <- keyword' "end"
>        return (group (pko <> hole <+> pke))
> ppFrame (New1 {--})
>   =  do
>        pk <- keyword "new"
>        parensPrec 10 $
>          return (group (pk <> hole))
> ppFrame (Append1 {--} e env)
>   =  do
>        pe <- withPrec 11 $ ppExpr e
>        penv <- ppEnv_ (free e) env
>        return (group (hole <+> pe <+> penv))
> ppFrame (Append2 v {--})
>   =  do
>        pv <- withPrec 11 $ ppVal v
>        return (group (pv <+> hole))
> ppFrame (Invoke1 {--} l e2 _)
>   =  parensPrec 10 $ do
>        pe2 <- withPrec 11 $ ppExpr e2
>        -- what about the env?
>        return (group (hole <> text "." <> ppLab l <> text "~" <> pe2))
> ppFrame (Invoke2 v1 l {--})
>   =  parensPrec 10 $ do
>        pv1 <- withPrec 11 $ ppVal v1
>        return (group (pv1 <> text "." <> ppLab l <> text "~" <> hole))

> ppFrame (Inherit1 {--})
>   =  do
>        pk <- keyword "inherit"
>        return (group (pk <> hole))
> ppFrame (InheritIn1 x {--} e2 env)
>   =  do
>        pkl <- keyword "inherit"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pe2 <- withPrec 0 $ ppExpr e2
>        penv <- ppEnv_ (difference (free e2) (singleton x)) env
>        return (group (pkl <> block 8 (text x <> text " =" <+> hole) <> 
>                       block 1 (sp <> pki <> block 3 pe2) <+> pke <+> penv))
> ppFrame (InheritIn2 x {--} e2 env)
>   =  do
>        pkl <- keyword "inherit"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pe2 <- withPrec 0 $ ppExpr e2
>        penv <- ppEnv_ (difference (free e2) (singleton x)) env
>        return (group (pkl <> block 8 (text x <> text " =" <+> hole) <> 
>                       block 1 (sp <> pki <> block 3 pe2) <+> pke <+> penv))
> ppFrame (InheritIn3 x v1 {--})
>   =  do
>        pkl <- keyword "inherit"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pv1 <- withPrec 0 $ ppVal v1
>        return (group (pkl <> block 8 (text x <> text " =" <+> pv1) <> 
>                       block 1 (sp <> pki <> block 3 hole) <+> pke))
> ppFrame (Include1 {--})
>   =  do
>        pk <- keyword "include"
>        return (group (pk <> hole))
> ppFrame (IncludeIn1 x {--} e2 env)
>   =  do
>        pkl <- keyword "include"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pe2 <- withPrec 0 $ ppExpr e2
>        penv <- ppEnv_ (difference (free e2) (singleton x)) env
>        return (group (pkl <> block 8 (text x <> text " =" <+> hole) <> 
>                       block 1 (sp <> pki <> block 3 pe2) <+> pke <+> penv))
> ppFrame (IncludeIn2 x v1 {--})
>   =  do
>        pkl <- keyword "include"
>        pki <- keyword "in"
>        pke <- keyword' "end"
>        pv1 <- withPrec 0 $ ppVal v1
>        return (group (pkl <> block 8 (text x <> text " =" <+> pv1) <> 
>                       block 1 (sp <> pki <> block 3 hole) <+> pke))

> ppFrame (ArrFun1 {--} x e2 env)
>   =  parensPrec 10 $ do
>        pe2 <- withPrec 0 $ ppExpr e2
>        penv <- ppEnv_ (difference (free e2) (singleton x)) env
>        pa <- alignkey "array" (brackets hole <+> text x <> text " =>" <+> pe2)
>        return (pa <+> penv)
> ppFrame (ArrFun2i i n vs x e2 env)
>   =  parensPrec 10 $ do
>        pvs <- mapM ppVal vs
>        pe2 <- withPrec 0 $ ppExpr e2
>        penv <- ppEnv_ (difference (free e2) (singleton x)) env
>        pa <- alignkey "array" (text ("[" ++ show i ++ ".." ++ show (n-1) ++ "]") <+> text x <> text " =>" <+> pe2)
>        return (brackets (intersperse (char ',' <> sp) (pvs ++ [hole])) <+> text "+" <+>
>                pa <+> penv)
> ppFrame (ArrListi vs es env)
>   =  do
>        pvs <- mapM ppVal vs
>        pes <- mapM ppExpr es
>        penv <- ppEnv_ (unions (map free es)) env
>        return (brackets (intersperse (char ',' <> sp) (pvs ++ [hole] ++ pes) <+> penv))
> ppFrame (ArrListn vs)
>   =  do
>        pvs <- mapM ppVal vs
>        return (brackets (intersperse (char ',' <> sp) (pvs ++ [hole])))
> ppFrame (Subscript1 {--} e2 env)
>   =  parensPrec 10 $ do
>        pe2 <- withPrec 0 $ ppExpr e2
>        penv <- ppEnv_ (free e2) env
>        return (group (hole <> text "." <> brackets pe2 <+> penv))
> ppFrame (Subscript2 v1 {--})
>   =  parensPrec 10 $ do
>        pv1 <- withPrec 11 $ ppVal v1
>        return (group (pv1 <> text "." <> brackets hole))

 ppFrame (Force1 {--})
   =  group (text "force " <> hole)

 ppFrame f -- catch-all (vorlufig)
   =  pretty f

> recordi :: [Row Val] -> Row () -> [Row Expr] -> Env -> Frame
> recordi fs1 l []  _    =  Recordn fs1 l
> recordi fs1 l fs2 env  =  Recordi fs1 l fs2 env

> arrFun2i :: Integer -> Integer -> [Val] -> Var -> Expr -> Env -> Frame
> arrFun2i i n vs x e2 env
>   | i < n      =  ArrFun2i i n vs x e2 env
>   | otherwise  =  ArrListn vs

> arrListi :: [Val] -> [Expr] -> Env -> Frame
> arrListi vs [] _    =  ArrListn vs
> arrListi vs es env  =  ArrListi vs es env

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Memory}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type Loc
>   =  Integer
> type Mem
>   =  (Loc, [(Loc, Val)]) -- nchste freie Location

> ppMem :: Mem -> P Doc
> ppMem (_, mem)
>   =  ppList ppCell mem

> ppCell :: (Loc, Val) -> P Doc
> ppCell (a, v)
>   =  do
>        pv <- withPrec 0 $ ppVal v
>        return (group (text "&" <> pretty a <> text ":" <+> pv))

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Instructions}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data State
>   =   Stop Val Mem
>   |   Forward  Expr Env Stack Mem
>   |   Backward Val Stack Mem
>   |   CoreDumped Doc

> {-# INLINE xforward #-}
> xforward :: Expr -> Env -> Stack -> Mem -> IO State
> xforward e env ctl mem
>   =  return (Forward e env ctl mem)
> {-# INLINE xbackward #-}
> xbackward :: Val -> Stack -> Mem -> IO State
> xbackward v ctl mem
>   =  return (Backward v ctl mem)
> {-# INLINE coreDumped #-}
> coreDumped :: Doc -> IO State
> coreDumped s
>   =  return (CoreDumped s)

> {-# INLINE runtimeTypeError #-}
> runtimeTypeError :: Options -> String -> String -> Val -> IO State
> runtimeTypeError o name tp arg
>   =  coreDumped (text "`" <> text name <> text "' expects " <> text tp <> text ":" <> nl <>
>        (group (text name <+> pVal o 11 arg)))

> {-# INLINE throwException #-}
> throwException :: Con -> [Row Val] -> [Frame] -> Mem -> IO State
> throwException x xs ctl mem
>   =  xbackward (VCon x (VRecord xs)) (Throw1 {--} : ctl) mem

> instance Show State where
>   showsPrec _ (Stop v mem) = shows v . showString " | " . shows mem
>   showsPrec _ (Forward  e env ctl mem) 
>     = shows e . showString " " . shows env . showString " >> " . shows ctl . showString " | " . shows mem
>   showsPrec _ (Backward v ctl mem) = shows v . showString " << " . shows ctl . showString " | " . shows mem
>   showsPrec _ (CoreDumped d)  = showString ("core dumped: " ++ render (Page 80) d)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{The abstract machine}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

One step of the abstract machine.

> {-# INLINE step #-}
> step :: Options -> State -> IO State

> step _ (Stop v mem)
>   =  return (Stop v mem)
> step o (Forward e env ctl mem)
>   =  forward o e env ctl mem
> step _ (Backward v [] mem)
>   =  return (Stop v mem)
> step o (Backward v (f : fs) mem)
>   =  backward o f v fs mem
> step _ (CoreDumped s)
>   =  coreDumped s

Forward step.

> forward :: Options -> Expr -> Env -> Stack -> Mem -> IO State
> forward o (Bool b) _ ctl mem
>   =  xbackward (VBool b) ctl mem
> forward o (If e1 e2 e3) env ctl mem
>   =  xforward e1 env (If1 {--} e2 e3 env : ctl) mem
> forward o (Nat n) _ ctl mem
>   =  xbackward (VNat n) ctl mem
> forward o (UnOp op e) env ctl mem
>   =  xforward e env (UnOp1 op {--} : ctl) mem
> forward o (BinOp e1 op e2) env ctl mem
>   =  xforward e1 env (BinOp1 {--} op e2 env : ctl) mem
> forward o (String s) _ ctl mem
>   =  xbackward (VString s) ctl mem
> forward o Type _ ctl mem
>   =  xbackward VType ctl mem
> forward o (Var x) env ctl mem
>   =  case lookup x env of
>      Nothing  ->  coreDumped (text "unbound identifier: " <> text x)
>      Just v   ->  xbackward v ctl mem
> forward o (Let e1 e2) env ctl mem
>   =  xforward e1 env (Let1 {--} e2 env : ctl) mem
> forward o (Val p e) env ctl mem
>   =  xforward e env (Val1 p {--} : ctl) mem
> forward o (Fun x e) env ctl mem
>   =  xbackward (Closure x e False env) ctl mem
> forward o (App e1 e2) env ctl mem
>   =  xforward e1 env (App1 {--} e2 env : ctl) mem
> forward o (Rec p e) env ctl mem
>   =  do
>        let pvs = patvars p
>        rs <- mapM (\p' -> fmap (\r -> (p', VHole r)) (newIORef Nothing)) (reverse pvs) -- reverse to maintain correct order
>        let env' = Env rs \+/ env
>        xforward e env' (Rec1 p {--} (Env rs) : ctl) mem
> forward o (Sel e l) env ctl mem
>   =  xforward e env (Sel1 {--} l env : ctl) mem
> forward o (Record []) _ ctl mem
>   =  xbackward (VRecord []) ctl mem
> forward o (Record (f : fs)) env ctl mem
>   =  xforward e env (recordi [] r fs env : ctl) mem
>      where (r, e) = split f
> forward o (Con c e) env ctl mem
>   =  xforward e env (Con1 c {--} : ctl) mem
> forward o (Seq e1 e2) env ctl mem
>   =  xforward e1 env (Seq1 {--} e2 env : ctl) mem
> forward o (Ref e) env ctl mem
>   =  xforward e env (Ref1 {--} : ctl) mem
> forward o (Assign e1 e2) env ctl mem
>   =  xforward e1 env (Assign1 {--} e2 env : ctl) mem
> forward o (Deref e) env ctl mem
>   =  xforward e env (Deref1 {--} : ctl) mem
> forward o (LetCC x e) env ctl mem
>   =  xforward e (single (x, Cont ctl) \+/ env) ctl mem
> forward o (Case e rs) env ctl mem
>   =  xforward e env (Case1 {--} rs env : ctl) mem
> forward o (Continue e1 e2) env ctl mem
>   =  xforward e1 env (Continue1 {--} e2 env : ctl) mem
> forward o (Throw e) env ctl mem
>   =  xforward e env (Throw1 {--} : ctl) mem
> forward o (Try e rs) env ctl mem
>   =  xforward e env (Try1 {--} rs env : ctl) mem
> --forward (Write e) env ctl mem
> --  =  xforward e env (Write1 {--} : ctl) mem
> forward o (Call f e) env ctl mem
>   =  xforward e env (Call1 f {--} : ctl) mem
> forward o (While e1 e2) env ctl mem
>   =  xforward e1 env (While1 e1 e2 env : ctl) mem

> forward o (Comma e1 e2) env ctl mem
>   =  xforward e1 env (Comma1 {--} e2 env : ctl) mem
> forward o (Lookup e x) env ctl mem
>   =  xforward e env (Lookup1 {--} x : ctl) mem

> forward o (Class e) env ctl mem
>   =  xbackward (VClass e env) ctl mem
> forward o (Object e) env ctl mem
>   =  xforward e env (Object1 {--} : ctl) mem
> forward o (New e) env ctl mem
>   =  xforward e env (New1 {--} : ctl) mem
> forward o (Invoke e1 l e2) env ctl mem
>   =  xforward e1 env (Invoke1 {--} l e2 env : ctl) mem
> forward o Empty _ ctl mem
>   =  xbackward (VEnv empty) ctl mem
> forward o (Append e1 e2) env ctl mem
>   =  xforward e1 env (Append1 {--} e2 env : ctl) mem
> forward o (Method x e) env ctl mem
>   =  xbackward (VEnv (single (x, VMethod e env))) ctl mem
> forward o (Inherit e) env ctl mem
>   =  xforward e env (Inherit1 {--} : ctl) mem
> forward o (InheritIn x e1 e2) env ctl mem
>   =  xforward e1 env (InheritIn1 x {--} e2 env : ctl) mem
> forward o (Include e) env ctl mem
>   =  xforward e env (Include1 {--} : ctl) mem
> forward o (IncludeIn x e1 e2) env ctl mem
>   =  xforward e1 env (IncludeIn1 x {--} e2 env : ctl) mem

> forward o (ArrFun e1 x e2) env ctl mem
>   =  xforward e1 env (ArrFun1 {--} x e2 env : ctl) mem
> forward o (ArrList []) _ ctl mem
>   =  xbackward (VArray (listArray (0, -1) [])) ctl mem
> forward o (ArrList (e : es)) env ctl mem
>   =  xforward e env (arrListi [] es env : ctl) mem
> forward o (Subscript e1 e2) env ctl mem
>   =  xforward e1 env (Subscript1 {--} e2 env : ctl) mem
> forward o e env ctl mem
>   =  error (show e)

> {-

Tying the recursive knot (returns `Nothing' if `e' is not a value).

> knot :: Expr -> Env -> Maybe Val
> -- knot e _
> --   |  Debug.Trace.trace ("knot: " ++ show e) False = undefined
> knot (Bool b) _
>   =  Just (VBool b)
> knot (Nat n) _
>   =  Just (VNat n)
> knot (String s) _
>   =  Just (VString s)
> knot Type _
>   =  Just VType
> knot (Fun p e) env
>   =  Just (Closure p e True env)
> knot (Record rs) env
>   =  do  rs' <- mapM (\ r -> knotRow r env) rs
>          return (vrecord rs')
> knot (Con c e) env
>   =  do  v <- knot e env
>          return (VCon c v)

begin{experimentell}
Just an ordinary call-by-name interpreter (denotational semantics) for
a pure lambda-calculus.

> --knot (Var x) env
> --  =  lookup x env -- mglicherweise ein Blackhole, das Haskell entdeckt
> knot (App e1 e2) env
>   =  do  Closure p' e' _ env' <- knot e1 env
>          v2 <- knot e2 env
>          -- in the calculation of |inserttp|, it's important that |v2|
>          -- is only evaluated if we already know that we expect a type
>          -- argument ...
>          let rec p0 e0 env0 =
>               do  let inserttp  =  case p0 of
>                                      PatType  ->  case v2 of
>                                                     VType  ->  False
>                                                     _      ->  True
>                                      _              ->  False
>                   if inserttp
>                     then  do  Closure p'' e'' _ env'' <- knot e0 env0
>                               rec p'' e'' env''
>                               -- call recursively, we might need to insert another one
>                     else  knot e0 (Env (match' p0 v2) \+/ env0) -- TODO: der Musterabgleich kann fehlschlagen!!
>          rec p' e' env'
> knot (Sel e l) env
>   =  do  VRecord rs <- knot e env
>          select l rs

end{experimentell}

> knot _ _
>   =  Nothing
>
> knotRow :: Row Expr -> Env -> Maybe (Row Val)
> knotRow (Named s e) env
>   =  do  v <- knot e env
>          return (Named s v)
> knotRow (Numbered e) env
>   =  do  v <- knot e env
>          return (Numbered v)

> -}

> split :: Row a -> (Row (), a)
> split (Named l a)  = (Named l (), a)
> split (Numbered a) = (Numbered (), a)

Backward step (return). Here all the dynamic checks are happening.

> backward :: Options -> Frame -> Val -> Stack -> Mem -> IO State

> backward o c (VHole r) ctl mem
>   =  do
>        v' <- readIORef r
>        case v' of
>          Nothing  -> throwException "Loop" [] ctl mem
>          Just v'' -> backward o c v'' ctl mem

> backward o (If1 {--} e2 e3 env) (VBool b) ctl mem
>   =  if b then xforward e2 env ctl mem
>           else xforward e3 env ctl mem
> backward o (If1 {--} _ _ _) _ _ _
>   =  coreDumped (text "`if' expects a Bool") -- TODO
> backward o (UnOp1 op {--}) (VNat v) ctl mem
>   =  xbackward (unop op v) ctl mem
> backward o (UnOp1 op {--}) v _ _
>   =  coreDumped (text "`" <> pretty op <> text "' expects a natural number:" <> nl <>
>        (group (pretty op <+> pVal o 11 v)))
> backward o (BinOp1 {--} op e2 env) v1 ctl mem
>   =  xforward e2 env (BinOp2 v1 op {--} : ctl) mem

NEW: |div| and |mod| raise an exception.

> backward o (BinOp2 (VNat _) Div {--}) (VNat 0) ctl mem
>   =  throwException "Div" [] ctl mem
> backward o (BinOp2 (VNat _) Mod {--}) (VNat 0) ctl mem
>   =  throwException "Mod" [] ctl mem

> backward o (BinOp2 (VNat v1) op {--}) (VNat v2) ctl mem
>   |  op /= Cat
>   =  xbackward (binopNat v1 op v2) ctl mem
> backward o (BinOp2 (VString v1) op {--}) (VString v2) ctl mem
>   |  not (op `elem` [Sub, Mul, Div, Mod])
>   =  xbackward (binopString v1 op v2) ctl mem
> backward o (BinOp2 v1 op {--}) v2 _ _
>   =  let optext | op == Cat = "two strings"
>                 | op `elem` [Sub, Mul, Div, Mod] = "two natural numbers"
>                 | otherwise = "two natural numbers or two strings"
>      in  coreDumped (text "`" <> pretty op <> text "' expects " <> text optext <> text ":" <> nl <>
>        (group (pVal o 11 v1 <+> pretty op <+> pVal o 11 v2)))

> backward o (Let1 {--} e2 env) (VEnv env1) ctl mem
>   =  xforward e2 (env1 \+/ env) ctl mem
> backward o (Val1 p {--}) v ctl mem
>   =  case match p v of
>      Nothing  -> throwException "Match" [] ctl mem
>      Just vne -> xbackward (VEnv (reverseCat vne empty)) ctl mem -- TODO: match sollte eine Umgebung liefern, keine Liste

> backward o (App1 {--} e2 env) v1@(Closure _ _ _ _) ctl mem  -- Funktionsaufruf
>   =  xforward e2 env (App2 v1 {--} : ctl) mem

 backward (App1 {--} e2 env) v1@(VMethod _ _ _) ctl mem    -- Methodenaufruf
   =  xforward e2 env (MApp2 v1 {--} env : ctl) mem

> backward o (App1 {--} e2 _) v1 _ _
>   =  coreDumped (text "`@': function or method expected:" <> nl <>
>        (group (pVal o 10 v1 <> text " @" <+> pExpr o 11 e2))) -- TODO: env fehlt

> backward o (App2 (Closure PatType e _ env) {--}) v2 ctl mem
>   |  case v2 of { VType -> False; _ -> True }  -- insert missing type argument
>   =  xforward e env (TApp2 {--} v2 : ctl) mem
> backward o (App2 (Closure p e _ env) {--}) v2 ctl mem
> --  =  xforward e (match' p v2 \+/ env) ctl mem
>   =  case match p v2 of
>      Nothing  -> throwException "Match" [] ctl mem
>      Just vne -> xforward e (reverseCat vne env) ctl mem
> backward o (App2 _ {--}) _ _ _
>   =  error "internal error in function application"
> backward o (TApp2 {--} v2) (Closure PatType e _ env) ctl mem
>   |  case v2 of { VType -> False; _ -> True }  -- insert missing type argument
>   =  xforward e env (TApp2 {--} v2 : ctl) mem
> backward o (TApp2 {--} v2) (Closure p e _ env) ctl mem
> --  =  xforward e (match' p v2 \+/ env) ctl mem
>   =  case match p v2 of
>      Nothing  -> throwException "Match" [] ctl mem
>      Just vne -> xforward e (reverseCat vne env) ctl mem
> backward o (TApp2 {--} v2) v1 _ _
>   =  coreDumped (text "`@': function or method expected:" <> nl <>
>        (group (pVal o 10 v1 <> text " @" <+> pVal o 11 v2))) -- TODO: env fehlt (??)

 backward (MApp2 (VMethod p e env) {--} env') v2 ctl mem
   =  xforward e (("self", fromJust (lookup "self" env')) : match' p v2 \+/ env) ctl mem

Direkter Methodenaufruf in der Klassendefinition (m e) ohne Qualifier
(c.m() oder self.m()): `m e' ist eine Abkrzung fr `self.m e'.

> backward o (Sel1 {--} l _) (VRecord v) ctl mem
>   =  case select l v of
>      Nothing -> coreDumped (text "missing field: " <> pretty l)
>      Just v'  -> xbackward v' ctl mem

> backward o (Sel1 {--} (AlNumLab l) _) v@(VObject env) ctl mem
>   =  case lookup l env of
>      Nothing  ->  coreDumped (text "message not understood: " <> text l)
>      Just (VMethod e env') ->  xforward e (single ("self", v) \+/ env') ctl mem
>      Just _ -> error "internal error in method selection"
>--      Just v'  -> xbackward v' ctl mem

Methoden werden speziell behandelt: `self' wird an das Objekt selbst
gebunden.  In allen anderen Fllen entspricht die Selektion der
Record-Selektion.

> backward o (Sel1 {--} (AlNumLab l) env') (VSuper env) ctl mem
>   =  case lookup l env of
>      Nothing  ->  coreDumped (text "message not understood: " <> text l)
>      Just (VMethod e env'') ->  xforward e (single ("self", fromJust (lookup "self" env')) \+/ env'') ctl mem
>      Just _ -> error "internal error in method selection"
>--      Just (VMethod p e env) ->  xbackward (Closure p e False (("self", ) : env)) ctl mem
>--      Just v'  -> xbackward v' ctl mem

Bei einem `super'-Aufruf wird nicht das Objekt selbst an `self'
gebunden, sondern die gegenwrtige Bindung von `self' weitergereicht.

> backward o (Sel1 {--} l _) v _ _
>   =  coreDumped (text "`.': record or object expected:" <> nl <>
>        (group (pVal o 11 v <> text " ." <+> ppLab l)))

 backward o (Force1 {--}) (Method e env) ctl mem
   =  xforward e env ctl mem

> backward o (Rec1 p {--} (Env rs)) v ctl mem
>   =  case match p v of
>      Nothing  -> throwException "Match" [] ctl mem
>      Just vne -> do
>                    -- putStrLn ("updating:" ++ show vne)
>                    -- update the environment
>                    mapM (\ (x, VHole r) -> case Prelude.lookup x (reverse vne) of
>                                              Nothing -> return () -- shouldn't happen, exception?
>                                              Just v' -> writeIORef r (Just v') ) rs
>                    xbackward v ctl mem

> backward o (Recordi fs1 r1 (f2 : fs2) env) v ctl mem
>   =  xforward e2 env (recordi (fs1 ++ [join r1 v]) r2 fs2 env : ctl) mem
>      where (r2, e2) = split f2
> backward o (Recordi _ _ _ _) _ _ _
>   =  error "internal error in record construction"
> backward o (Recordn fs l) v ctl mem
>   =  xbackward (vrecord (fs ++ [join l v])) ctl mem

---

> backward o (Con1 c {--}) v ctl mem
>   =  xbackward (VCon c v) ctl mem

> backward o (Seq1 {--} e2 env) _ ctl mem
>   =  xforward e2 env ctl mem
> backward o (Ref1 {--}) v ctl (n, mem)
>   =  xbackward (Addr n) ctl (n + 1, (n, v) : mem)
> backward o (Assign1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Assign2 v1 {--} : ctl) mem
> backward o (Assign2 (Addr a1) {--}) v2 ctl (n, mem)
>   =  xbackward (VRecord []) ctl (n, update a1 v2 mem)

> backward o (Assign2 (VRecord as) {--}) (VRecord vs) ctl (n, mem)
>   |  isJust avs
>   =  xbackward (VRecord []) ctl (n, foldl (\ m f -> f m) mem [ update a v | (a, v) <- fromJust avs ])
>      where avs = zipRecord as vs

> backward o (Assign2 _ {--}) _ _ _
>   =  coreDumped (text "`:=': address or tuple of addresses expected") -- TODO

> backward o (Deref1 {--}) (Addr a) ctl mem@(_, mem')
>   =  case Prelude.lookup a mem' of
>      Nothing -> coreDumped (text "segmentation violation: &" <> pretty a)
>      Just v  -> xbackward v ctl mem
> backward o (Deref1 {--}) _ _ _
>   =  coreDumped (text "`!': address expected") -- TODO

> backward o (Case1 {--} rs env) v ctl mem
>   =  case msum [ match p v >>= \ env' -> return (env', e) | (p, e) <- rs ] of
> --     Nothing        -> coreDumped (text "missing case") -- TODO
>      Nothing        -> throwException "Match" [] ctl mem
>      Just (env', e) -> xforward e (Env env' \+/ env) ctl mem
> backward o (Continue1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Continue2 v1 {--} : ctl) mem
> backward o (Continue2 (Cont ctl1) {--}) v2 _ mem
>   =  xbackward v2 ctl1 mem
> backward o (Continue2 v {--}) _ _ _
>   =  runtimeTypeError o "continue" "a continuation" v
> backward o (Throw1 {--}) v (Try1 {--} rs env : ctl) mem
>   =  case msum [ match p v >>= \ env' -> return (env', e) | (p, e) <- rs ] of
>      Nothing        -> xbackward v (Throw1 {--} : ctl) mem
>      Just (env', e) -> xforward e (Env env' \+/ env) ctl mem
> backward o (Throw1 {--}) (VCon c v1) (_ : ctl) mem
> --  =  backward (Throw1 {--}) (VCon c v1) ctl mem
>   =  xbackward (VCon c v1) (Throw1 {--} : ctl) mem
> backward o (Throw1 {--}) (VCon c v1) [] _
>   =  coreDumped (text "uncaught exception: " <> pVal o 0 (VCon c v1))
> backward o (Throw1 {--}) v _ _
>   =  runtimeTypeError o "throw" "an exception" v
> backward o (Try1 {--} _ _) v ctl mem
>   =  xbackward v ctl mem

TODO: Fehlerbehandlung Call1 (a la contracts?!)

> backward o (Call1 Chr {--}) (VNat n) ctl mem
>   =  xbackward (VString [chr (fromInteger n `mod` 128)]) ctl mem
> backward o (Call1 Chr {--}) v _ _
>   =  runtimeTypeError o "chr" "a natural number" v
> backward o (Call1 GetChar {--}) (VRecord []) ctl mem
>   =  catch  (getChar >>= \ c -> xbackward (VString [c]) ctl mem)
>             (\ _ -> throwException "EOF" [] ctl mem)
> backward o (Call1 GetChar {--}) v _ _
>   =  runtimeTypeError o "getchar" "a unit" v
> backward o (Call1 ReadLine {--}) (VString prompt) ctl mem
>   =  Readline.readline prompt >>= \ inp ->
>         case inp of
>           Nothing -> throwException "EOF" [] ctl mem
>           Just s  -> do when (not (null s)) (Readline.addHistory s)
>                         xbackward (VString s) ctl mem
> backward o (Call1 ReadLine {--}) v _ _
>   =  runtimeTypeError o "read-line" "a string" v
> backward o (Call1 Head {--}) (VString s) ctl mem
>   =  xbackward (VString (case s of "" -> ""; c : _ -> [c])) ctl mem
> backward o (Call1 Head {--}) v _ _
>   =  runtimeTypeError o "head" "a string" v
> backward o (Call1 Null {--}) (VString s) ctl mem
>   =  xbackward (VBool (null s)) ctl mem
> backward o (Call1 Null {--}) v _ _
>   =  runtimeTypeError o "null" "a string" v

> backward o (Call1 Ord {--}) (VString []) ctl mem
>   =  throwException "Ord" [] ctl mem
> backward o (Call1 Ord {--}) (VString (c : _)) ctl mem
>   =  xbackward (VNat (toInteger (ord c))) ctl mem
> backward o (Call1 Ord {--}) v _ _
>   =  runtimeTypeError o "ord" "a string" v
> backward o (Call1 PutChar {--}) (VString s) ctl mem
>   =  (if null s then return () else putChar (head s)) >> xbackward (VRecord []) ctl mem
> backward o (Call1 PutChar {--}) v _ _
>   =  runtimeTypeError o "putchar" "a string" v
> backward o (Call1 Show {--}) v ctl mem
>   =  xbackward (VString (render OneLine (pVal o 0 v))) ctl mem

> backward o (Call1 Tail {--}) (VString s) ctl mem
>   =  xbackward (VString (case s of "" -> ""; _ : s' -> s')) ctl mem
> backward o (Call1 Tail {--}) v _ _
>   =  runtimeTypeError o "tail" "a string" v

> backward o (Call1 Write {--}) (VString s) ctl mem
>   =  putStr s >> xbackward (VRecord []) ctl mem
> backward o (Call1 Write {--}) v _ _
>   =  runtimeTypeError o "write" "a string" v

> backward o (Call1 Size {--}) (VArray a) ctl mem
>   =  xbackward (VNat (snd (bounds a) + 1)) ctl mem
> backward o (Call1 Size {--}) v _ _
>   =  runtimeTypeError o "size" "an array" v
> backward o (Call1 ReadFromFile {--}) (VString s) ctl mem
>   =  readFile s >>= \ contents -> xbackward (VString contents) ctl mem
> backward o (Call1 ReadFromFile {--}) v _ _
>   =  runtimeTypeError o "read-from-file" "a string" v
> backward o (Call1 WriteToFile {--})
>   (VRecord [Numbered (VString filepath), Numbered (VString contents)]) ctl mem
>   =  writeFile filepath contents >> xbackward (VRecord []) ctl mem
> backward o (Call1 WriteToFile {--}) v _ _
>   =  runtimeTypeError o "write-to-file" "a pair of strings" v

> backward o (Call1 System {--}) (VString s) ctl mem
>   =  system s >> xbackward (VRecord []) ctl mem
> backward o (Call1 System {--}) v _ _
>   =  runtimeTypeError o "system" "a string" v

> backward o (While1 e1 e2 env) (VBool b) ctl mem
>   =  if b then xforward e2 env (Seq1 {--} (While e1 e2) env : ctl) mem
>           else xbackward (VRecord []) ctl mem
> backward o (While1 _ _ _) v _ _
>   =  runtimeTypeError o "while" "a boolean" v

 backward (Open1 {--}) (VStruct env) ctl mem
   =  xbackward (VEnv env) ctl mem

 backward (LetOpen1 {--} _ _) v _ _
   =  runtimeTypeError o "open" "a struct" v

 backward (Struct1 {--}) (VEnv env) ctl mem
   =  xbackward (VStruct env) ctl mem

> backward o (Comma1 {--} e2 env) v1@(VEnv env1) ctl mem
>   =  xforward e2 (env1 \+/ env) (Comma2 v1 {--} : ctl) mem
> backward o (Comma2 (VEnv env1) {--}) (VEnv env2) ctl mem
>   =  xbackward (VEnv (env2 \+/ env1)) ctl mem
> backward o (Comma2 _ {--}) _ _ _
>   =  error "internal error in environment construction"

> backward o (Lookup1 {--} x) (VEnv env) ctl mem
>   =  forward o (Var x) env ctl mem
> backward o (Lookup1 {--} _) _ _ _
>   =  error "internal error in lookup of qualified identifier"

 backward o (LetInherit1 Nothing {--} e2 env) (VObject env') ctl mem
   =  xforward e2 (env' \+/ env) ctl mem
 backward o (LetInherit1 (Just x) {--} e2 env) (VObject env') ctl mem
   =  xforward e2 (single (x, VSuper env') \+/ env' \+/ env) ctl mem

Wird die geerbte Klasse an einen (Werte-) Bezeichner gebunden, wird
die Klasse als `Super' markiert in der Umgebung abgelegt, sozusagen
als Unterobjekt.

> backward o (Object1 {--}) (VEnv env) ctl mem
>   =  xbackward (VObject env) ctl mem
> backward o (New1 {--}) (VClass e env) ctl mem
>   =  xforward e env (Object1 {--} : ctl) mem
> backward o (New1 {--}) v _ _
>   =  runtimeTypeError o "new" "a class" v

> backward o (Invoke1 {--} l e2 env) v1 ctl mem
>   =  xforward e2 env (Invoke2 v1 l {--} : ctl) mem
> backward o (Invoke2 (VObject env) (AlNumLab l) {--}) v2 ctl mem
>   =  case lookup l env of
>      Nothing  ->  coreDumped (text "message not understood: " <> text l)
>      Just (VMethod e env') ->  xforward e (single ("self", v2) \+/ env') ctl mem
>      Just _ -> error "internal error in method invocation"
> backward o (Invoke2 (VSuper env) (AlNumLab l) {--}) v2 ctl mem -- VObject und VSuper werden gleich behandelt
>   =  case lookup l env of
>      Nothing  ->  coreDumped (text "message not understood: " <> text l)
>      Just (VMethod e env') ->  xforward e (single ("self", v2) \+/ env') ctl mem
>      Just _ -> error "internal error in method invocation"
> backward o (Invoke2 v l {--}) _ _ _
>   =  coreDumped (text "`.': object expected:" <> nl <>
>        (group (pVal o 11 v <> text " ." <+> ppLab l)))

> backward o (Append1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Append2 v1 {--} : ctl) mem
> backward o (Append2 (VEnv env1) {--}) (VEnv env2) ctl mem
>   =  xbackward (VEnv (env2 \+/ env1)) ctl mem
> backward o (Append2 _ {--}) _ _ _
>   =  error "internal error in object construction"

> backward o (Inherit1 {--}) (VClass e env) ctl mem
>   =  xforward e env ctl mem
> backward o (Inherit1 {--}) v _ _
>   =  runtimeTypeError o "inherit" "a class" v

> backward o (InheritIn1 x {--} e2 env) (VClass e env') ctl mem
>   =  xforward e env' (InheritIn2 x {--} e2 env : ctl) mem
> backward o (InheritIn1 x {--} _ _) v _ _
>   =  coreDumped (text "`inherit': class expected:" <> nl <>
>        (group (text "inherit " <> text x <> text " =" <+> pVal o 0 v)))
> backward o (InheritIn2 x {--} e2 env) v@(VEnv env') ctl mem
>   =  xforward e2 (single (x, VSuper env') \+/ env) (InheritIn3 x v {--} : ctl) mem
> backward o (InheritIn2 _ {--} _ _) _ _ _
>   =  error "internal error in InheritIn2"
> backward o (InheritIn3 _ (VEnv env1) {--}) (VEnv env2) ctl mem
>   =  xbackward (VEnv (env2 \+/ env1)) ctl mem
> backward o (InheritIn3 _ _ {--}) _ _ _
>   =  error "internal error in InheritIn3"

> backward o (Include1 {--}) (VObject env) ctl mem
>   =  xbackward (VEnv env) ctl mem
> backward o (Include1 {--}) v _ _
>   =  runtimeTypeError o "include" "an object" v

> backward o (IncludeIn1 x {--} e2 env) v@(VObject env') ctl mem
>   =  xforward e2 (single (x, VSuper env') \+/ env) (IncludeIn2 x v {--} : ctl) mem
> backward o (IncludeIn1 x {--} _ _) v _ _
>   =  coreDumped (text "`include': object expected:" <> nl <>
>        (group (text "include " <> text x <> text " =" <+> pVal o 0 v)))
> backward o (IncludeIn2 _ (VObject env1) {--}) (VEnv env2) ctl mem
>   =  xbackward (VEnv (env2 \+/ env1)) ctl mem
> backward o (IncludeIn2 _ _ {--}) _ _ _
>   =  error "internal error in IncludeIn2"

> backward o (ArrFun1 {--} x e2 env) (VNat n) ctl mem
>   |  n == 0
>   =  xbackward (VArray (listArray (0, -1) [])) ctl mem
>   |  otherwise
>   =  xforward e2 (single (x, VNat 0) \+/ env) (arrFun2i 1 n [] x e2 env : ctl) mem
> backward o (ArrFun1 {--} _ _ _) v _ _
>   =  coreDumped (text "`array': natural number expected:" <> nl <>
>        (group (text "array [" <> pVal o 0 v <> text "]")))
> backward o (ArrFun2i i n vs x e2 env) v ctl mem
>   =  xforward e2 (single (x, VNat i) \+/ env) (arrFun2i (i + 1) n (vs ++ [v]) x e2 env : ctl) mem
> backward o (ArrListi vs (e : es) env) v ctl mem
>   =  xforward e env (arrListi (vs ++ [v]) es env : ctl) mem
> backward o (ArrListi _ _ _) _ _ _
>   =  error "internal error in array construction"
> backward o (ArrListn vs) v ctl mem
>   =  xbackward (VArray (listArray (0, genericLength vs) (vs ++ [v]))) ctl mem
> backward o (Subscript1 {--} e2 env) v1 ctl mem
>   =  xforward e2 env (Subscript2 v1 {--} : ctl) mem
> backward o (Subscript2 (VArray a) {--}) (VNat i) ctl mem
>   |  inRange (bounds a) i
>   =  xbackward (a!i) ctl mem
>   |  otherwise
>   =  throwException "Index" [] ctl mem
> backward o (Subscript2 _ {--}) _ _ _
>   =  coreDumped (text "array subscript expects an array and a Nat")

TODO: Wenn der Musterabgleich fehlschlgt: Match exception auslsen.
Als Zwischenloesung geben wir die leere Bindungsliste zurueck. Das
fuehrt dann normalerweise zu einem Core-Dump wegen eines ungebundenen
Bezeichners, das ist aber immer noch besser als ein
Haskell-Runtime-Error.

> match' :: Pat -> Val -> [(Var,Val)]
> match' p v = maybe [] reverse (match p v)

Beachte: die Bindungen werden gespiegelt, so dass spteren Bindungen
Vorrang eingerumt wird.

> match :: Pat -> Val -> Maybe [(Var,Val)]
> match PatFail _
>   =  Nothing
> match PatSucceed _
>   =  Just []
> match (PatVar x) v
>   =  Just [(x, v)]
> match (PatCon c p) (VCon c' v)
>   |  c == c'
>   =  match p v
> match PatType VType
>   =  Just []
> match (PatCon _ _) _
>   =  Nothing
> match (PatRecord ps) (VRecord vs)
>   =  do  envs <- matchRow ps vs
>          return (Prelude.concat envs)

> match (PatBoth p1 p2) v
>   =  do  env1 <- match p1 v
>          env2 <- match p2 v
>          return (env1 ++ env2)
> match (PatAlt p1 p2) v
>   =  match p1 v `mplus` match p2 v
> match _ _
>   =  Nothing

Annahme: Records sind sortiert/normiert.

> matchRow :: [Row Pat] -> [Row Val] -> Maybe [[(Var, Val)]]
> matchRow [] _         =  return []
> matchRow _ []         =  return []  -- eigentlich: Nothing
> matchRow (Named x p : pr) (Named y v : vr)
>   |  x == y           =  liftM2 (:) (match p v) (matchRow pr vr)
> matchRow (Numbered p : pr) (Numbered v : vr)
>                       =  liftM2 (:) (match p v) (matchRow pr vr)
> matchRow pr (_ : vr)  =  matchRow pr vr

> join :: Row a -> b -> Row b
> join (Named l _)  a = Named l a
> join (Numbered _) a = Numbered a

> select :: Lab -> [Row a] -> Maybe a
> select (AlNumLab l) = sel
>   where sel []                 =  Nothing
>         sel (Named l' a : fs)
>           | l == l'            =  Just a
>           | otherwise          =  sel fs
>         sel (Numbered _ : fs)  =  sel fs
> select (NumLab i)                      = sel (i - 1) -- Numerierung beginnt bei 1 (wg `fst' und `snd')
>   where sel n (Named _ _ : fs)         =  sel n fs
>         sel 0 (Numbered a : _)         =  Just a
>         sel (n + 1) (Numbered _ : fs)  =  sel n fs
>         sel _ _                        =  Nothing

> zipRecord :: [Row Val] -> [Row Val] -> Maybe [(Integer, Val)]
> zipRecord [] []
>   =  return []
> zipRecord (Named s1 (Addr a): as) (Named s2 v : vs)
>   |  s1 == s2  
>   =  do avs <- zipRecord as vs
>         return ((a, v) : avs)
> zipRecord (Numbered (Addr a) : as) (Numbered v : vs)
>   =  do avs <- zipRecord as vs
>         return ((a, v) : avs)
> zipRecord _ _
>   =  Nothing

> update :: (Eq a) => a -> t -> [(a,t)] -> [(a,t)]
> update _ _ [] = []
> update a v ((b, w) : bs)
>   | a == b    = (b, v) : bs
>   | otherwise = (b, w) : update a v bs

> patvars :: Pat -> [Var]
> patvars p =
>   case p of
>     PatFail -> []
>     PatSucceed -> []
>     PatVar v -> [v]
>     PatCon _ p' -> patvars p'
>     PatType -> []
>     PatRecord r -> concatMap patvarsRow r
>     PatBoth p1 p2 -> patvars p1 ++ patvars p2
>     PatAlt p1 p2 -> nub (patvars p1 ++ patvars p2)  -- both should be equal

> patvarsRow :: Row Pat -> [Var]
> patvarsRow (Named _ p) = patvars p
> patvarsRow (Numbered p) = patvars p

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Primitive operations}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

Unary and binary operators.

> unop :: UnOp -> Integer -> Val
> --unop Neg n  =  VNat (negate n) -- Hmm: fr natrliche Zahlen
> unop Neg n  =  VNat n

> binopNat :: Integer -> BinOp -> Integer -> Val
> binopNat n1 Add n2  =  VNat  (n1 +  n2)
> binopNat n1 Sub n2  =  VNat  ((n1 -  n2) `max` 0) -- spezielle Subtraktion auf den natrlichen Zahlen
> binopNat n1 Mul n2  =  VNat  (n1 *  n2)
> binopNat n1 Div n2  =  VNat  (n1 `div` n2)
> binopNat n1 Mod n2  =  VNat  (n1 `mod` n2)
> binopNat n1 Lt  n2  =  VBool (n1 <  n2)
> binopNat n1 Leq n2  =  VBool (n1 <= n2)
> binopNat n1 Eq  n2  =  VBool (n1 == n2)
> binopNat n1 Neq n2  =  VBool (n1 /= n2)
> binopNat n1 Geq n2  =  VBool (n1 >= n2)
> binopNat n1 Gt  n2  =  VBool (n1 >  n2)
> binopNat _  _   _   =  error "internal error in binopNat"

NB. `div' und 'mod' sind die sogenannten mathematischen Varianten (im
Unterschied zu `quot' und `rem'): (a + km) `mod` m = a `mod` m.

TODO: hier knnen Fehler auftreten: Division durch 0!

> binopString :: String -> BinOp -> String -> Val
> binopString s1 Cat s2  =  VString (s1 ++ s2)
> binopString s1 Add s2  =  VString (s1 ++ s2)
> binopString s1 Lt  s2  =  VBool (s1 <  s2)
> binopString s1 Leq s2  =  VBool (s1 <= s2)
> binopString s1 Eq  s2  =  VBool (s1 == s2)
> binopString s1 Neq s2  =  VBool (s1 /= s2)
> binopString s1 Geq s2  =  VBool (s1 >= s2)
> binopString s1 Gt  s2  =  VBool (s1 >  s2)
> binopString _  _   _   =  error "internal error in binopString"

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Initial environment}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

Die initiale Umgebung dient dazu, primitive Konstanten und Funktionen
wie `false' oder `write' bekanntzumachen, da diese nicht in
`System.bpl' definiert werden knnen. Alternativ: Bezeichner als
Schlsselwrter reservieren, so dass `write' stets die primitive
Funktion meint.

> closure :: Var -> Expr -> Val
> closure x e
>   =  Closure (PatVar x) e False empty

> stdenv', stdenv :: Env
> stdenv'
>   =  Env [("show",    closure "x" (Call Show (Var "x"))),
>           ("write",   closure "x" (Call Write (Var "x")))]
> stdenv
>   =  Env [("chr",     closure "x" (Call Chr (Var "x"))),
>           ("false",   VBool False),
>           ("get-char", closure "x" (Call GetChar (Var "x"))),
>           ("read-line", closure "x" (Call ReadLine (Var "x"))),
>           ("read-from-file", closure "x" (Call ReadFromFile (Var "x"))),
>           ("write-to-file", closure "x" (Call WriteToFile (Var "x"))),
>           ("system", closure "x" (Call System (Var "x"))),
>           ("head",    closure "x" (Call Head (Var "x"))),
>           ("null?",    closure "x" (Call Null (Var "x"))),
>           ("ord",     closure "x" (Call Ord (Var "x"))),
>           ("put-char", closure "x" (Call PutChar (Var "x"))),
>           ("show",    closure "x" (Call Show (Var "x"))),
>           ("size",    closure "x" (Call Size (Var "x"))),
>           ("tail",    closure "x" (Call Tail (Var "x"))),
>           ("true",    VBool True),
>           ("write",   closure "x" (Call Write (Var "x")))]

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Execute}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> execute :: Options -> Expr -> Env -> Mem -> IO State
> execute options e env mem
>   =  do catch
>           (do hSetBuffering stdout NoBuffering
>               hSetBuffering stdin  NoBuffering)
>           (\_ -> return ())
>         if trace options then
>           debug options (Forward e env [] mem)
>          else
>           exec options (Forward e env [] mem)

> exec :: Options -> State -> IO State
> exec _ cfg@(Stop _ _)
> --  =  putStrLn (render (Page 80) (ppState cfg)) >> 
>   =  return cfg
> exec _ cfg@(CoreDumped _)
> --  =  putStrLn (render (Page 80) (ppState cfg)) >> 
>   =  return cfg
> exec o cfg 
>   =  step o cfg >>= exec o

 initial options e
   =  if Bare `elem` options
        then Forward e [] [] (0, [])
        else Forward e stdenv [] (0, [])

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Tracer}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

 trace options e 
   = do hSetBuffering stdout NoBuffering
        hSetBuffering stdin  NoBuffering
        debug (initial options e)

> debug :: Options -> State -> IO State 
> debug o cfg
>   =  do  putStrLn (render (Page 80) (pState o cfg))

          putStr "? "
          inp <- getLine

>          let loop =
>               do
>                 inp <- Readline.readline "? "
>                 case inp of
>                   Nothing -> return cfg
>                   Just x ->
>                     if x `elem` [""] then
>                       step o cfg >>= debug o
>                     else if all isDigit x then
>                       apply (read x) (step o) cfg >>= debug o
>                     else if x `elem` [":q", ":quit"] then
>                       exec o cfg
>                     else 
>                       let v = lookup x (case cfg of Forward _ env _ _ -> env; _ -> empty)
>                       in  do
>                             case v of
>                               Just v'  ->  putStrLn (render (Page 80) (pVal o 0 v'))
>                               Nothing  ->  putStrLn (render (Page 80) (text ("unknown: " ++ x)))
>                             loop
>          loop

> pState :: Options -> State -> Doc
> pState o s = runReader (ppState s) (defaultPState { pcolor = O.color o, pdebug = O.debug o })

> ppState :: State -> P Doc
> ppState (Stop v _)
>   =  ppVal v
> --     <> nl <> text "mem: " <> block 5 (ppMem mem)
> ppState (Forward  e env ctl mem) 
>   =  do
>        pe <- ppExpr e
>        penv <- ppEnv_ (free e) env
>        ps <- ppStack ctl
>        pm <- ppMem mem
>        return $
>          text ">>>> " <> block 5 pe <> nl <>
>          text "env: " <> block 5 penv <> nl <>
>          text "ctl: " <> block 5 ps
>          <> nl <> text "mem: " <> block 5 pm
> ppState (Backward v ctl mem) 
>   =  do
>        pv <- ppVal v
>        ps <- ppStack ctl
>        pm <- ppMem mem
>        return $
>          text "<<<< " <> block 5 pv <> nl <>
>          text "ctl: " <> block 5 ps
>          <> nl <> text "mem: " <> block 5 pm
> ppState (CoreDumped d)  
>   =  return (text "core dumped: " <> d)

> ppStack :: Stack -> P Doc
> ppStack
>   =  ppList ppFrame

> apply :: Integer -> (a -> IO a) -> a -> IO a
> apply 0 _ x
>   =  return x
> apply (n + 1) f x
>   =  apply n f x >>= f

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Tests}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

 cfg = initial (BinOp (Nat 4711) Add (Nat 815))

debug cfg

 go e = debug (initial e)

> fun :: Var -> Expr -> Expr
> fun x e = Fun (PatVar x) e

go (fun "x" (Var "x"))
go (fun "x" (Var "x") $ Nat 4711)
go dd
go (factorial $ Num 3)
go (fromto $ Nat 1 $ Nat 9)

> dd :: Expr
> dd = fun "x" (Var "x" $$ Var "x") $$ fun "x" (Var "x" $$ Var "x") -- ;-)

> factorial :: Expr
> factorial
>   = Rec (PatVar "factorial") (fun "n" (
>       If (BinOp (Var "n") Eq (Nat 0)) 
>          (Nat 1) 
>          (BinOp (Var "n") Mul (Var "factorial" $$ BinOp (Var "n") Sub (Nat 1))) ))

> fromto :: Expr
> fromto
>   = Rec (PatVar "fromto") (fun "l" (fun "r" (
>       If (BinOp (Var "l") Gt (Var "r"))
>          (Con "Nil" (Record []))
>          (Con "Cons" (Record [ Named hd (Var "l"),
>                                Named tl ( Var "fromto" $$ BinOp (Var "l") Add (Nat 1) $$ Var "r")])) )))

> hd, tl :: String
> hd = "head"
> tl = "tail"
