%------------------------------------------------------------------------------
\section{Pretty printer for abstract syntax}
%------------------------------------------------------------------------------

> module Printer
> where
> import Control.Monad.Reader
>
> import Message hiding (And, break)
> import AbstractSyntax

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{The |P| monad}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

Pretty printing works in a reader monad and can be parameterized by the
precedence of the context and certain configuration flags.

> type P a = Reader PState a
> data PState = PState
>                 {  pprec   ::  !Int,
>                    pcolor  ::  !Bool, 
>                    pdebug  ::  !Bool
>                 }

> defaultPState = PState 0 False False 

> withPrec :: Int -> P a -> P a
> withPrec n = local (\ s -> s { pprec = n })

> parensPrec :: Int -> P Doc -> P Doc
> parensPrec m p = asks pprec >>= \ c -> if c > m then liftM parens p else p

> keyword, keyword' :: String -> P Doc
> keyword' x
>   =  do
>        pk <- ppCol askeyword 
>        return (pk (text x))
> keyword  x = keyword' x >>= \ px -> return (px <> text " ")

> ppCol :: (Doc -> Doc) -> P (Doc -> Doc)
> ppCol d =  asks pcolor >>= \ c -> if c then return d else return id

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Pretty printing: expressions}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

Names.

> pCon x = runReader (ppCon x) defaultPState 
> pLCon c d x = runReader (ppLCon x) (defaultPState { pcolor = c, pdebug = d })

> ppCon x 
>   =  do
>        pc <- ppCol ascon
>        return (pc (text x))
> ppLCon x 
>   =  do
>        pc <- ppCol ascon
>        return (pc (ltext x))

> ppRow :: (Type -> P Doc) -> (TyRow -> P Doc) 
> ppRow pp (TyField (Just s@(AlNumLab _)) Nothing a)
>   =  do
>        pa <- pp a
>        return (block 6 (ppLab s <> text " = " <> pa))
> ppRow pp (TyField _ _ a)
>   =  pp a -- assumes that numeric labels are always ordered correctly

TODO: add general pretty-printing for expressions.

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Pretty printing: patterns}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> ppPat :: Pat -> P Doc
> ppPat (PatTyped (PatVar (_,"_")) t1)  -- special construction in function types
>   =  ppType' t1
> ppPat (PatVar x)
>   =  return (ltext x)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Pretty printing: types}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> ppTyName :: LCon -> P Doc
> ppTyName c = asks pdebug >>= \ d -> if d then return (ltext c)  -- print data types including identification stamps
>                                          else (return . text . fst . splitTyName) c -- print data types without numbers

Operator precedence (from parser):
right 1 "->"
nonassoc 0 FORALL
right 1 "&"
right 0 "|"
right 2 "||"
nonassoc 1 ":="
right 0 ";"

> pType :: Bool -> Bool -> Type -> Doc
> pType c d t = runReader (ppType t) (defaultPState { pcolor = c, pdebug = d })

> {-
> parensType :: Type -> Doc
> parensType t = runReader (ppType t) (defaultPState { pprec = 11 })
> -}

> pType' :: Type -> Doc
> pType' t = runReader (ppType' t) defaultPState 

mit Farbe:

> ppType :: Type -> P Doc
> ppType t = do pc <- ppCol astype
>               d <- ppType' t
>               return (pc d)

ohne Farbe:

> ppType' :: Type -> P Doc
> ppType' (TyVar x)
>   =  ppTyName x
> ppType' (TyCon x)
>   =  ppTyName x
> -- ppType' (QualTyCon s x)
> --   =  ppStrExpr s <> text "." <> text x
> ppType' (TyApp t [])
>   =  ppType' t
> ppType' (TyApp t ts)
>   =  parensPrec 10 $ do
>        pt  <- withPrec 10 $ ppType' t
>        pts <- withPrec 11 $ mapM ppActual ts 
>        return (group (strip (pt : pts)))
> ppType' (TyAnd t1 t2)
>   =  parensPrec 1 $ do
>        pt1 <- withPrec 2 $ ppType' t1
>        pt2 <- withPrec 1 $ ppType' t2
>        return (group (pt1 <> text " &" <+> pt2))
> ppType' (TyFun [] t)
>   =  ppType' t
> {-
> ppType' (TyFun ts t)
>   =  parens (group (strip (map ppFormal ts)  <> text " ->" <+> ppType' t))
> -}
> -- Ralf suggests to pretty-print all functions with multiple arguments in a
> -- curried style:
> ppType' (TyFun (t:ts) t')
>   =  parensPrec 1 $ do
>        pk  <- keyword "fun"
>        pt  <- withPrec 2 $ ppFormal t
>        pt' <- withPrec 1 $ ppType' (TyFun ts t')
>        return (group (pk <> pt <> text " " <> block 2 (text "->" <+> pt')))
> ppType' (TyForall [] t)
>   =  ppType' t

NEU:

 ppType' (TyForall ts t)
   =  parens (group (keyword "forall" <> strip (map ppFormal ts) <> text " =>" <+> ppType' t))

> ppType' (TyForall (t:ts) t')
>   =  parensPrec 1 $ do
>        pt  <- withPrec 2 $ ppFormal t
>        pt' <- withPrec 1 $ ppType' (TyForall ts t')
>        return (group (pt <> text " " <> block 2 (text "->" <+> pt')))
> ppType' (TyRecord rs Nothing)
>   =  do
>        prs <- withPrec 0 $ mapM (ppRow ppType') rs
>        return (parens (group (intersperse (char ',' <> sp) prs)))  -- preliminary
> ppType' (TyRecord rs (Just n))
>   =  do
>        prs <- withPrec 0 $ mapM (ppRow ppType') rs
>        return (parens (group (intersperse (char ',' <> sp) (prs ++ [text ("_" ++ show n)]))))  -- preliminary
> ppType' (TyClass ss Nothing)
>   =  do
>        pm <- keyword "method"
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType' t >>= \ pt -> return $ pm <> block 5 (ltext x <> text " :" <+> pt)) ss
>        pk <- keyword' "class"
>        pke <- keyword' "end"
>        return (group (nest 3 (pk <> strip (empty : pss)) <+> pke))
> ppType' (TyClass ss (Just n))
>   =  do
>        pm <- keyword "method"
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType' t >>= \ pt -> return $ pm <> block 5 (ltext x <> text " :" <+> pt)) ss
>        pk <- keyword' "class"
>        pke <- keyword' "end"
>        return (group (nest 3 (pk <> strip (empty : pss ++ [text ("_" ++ show n)])) <+> pke))
> ppType' (TyObject ss Nothing)
>   =  do
>        pm <- keyword "method"
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType' t >>= \ pt -> return $ pm <> block 5 (ltext x <> text " :" <+> pt)) ss
>        pk <- keyword' "object"
>        pke <- keyword' "end"
>        return (group (nest 3 (pk <> strip (empty : pss)) <+> pke))
> ppType' (TyObject ss (Just n))
>   =  do
>        pm <- keyword "method"
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType' t >>= \ pt -> return $ pm <> block 5 (ltext x <> text " :" <+> pt)) ss
>        pk <- keyword' "object"
>        pke <- keyword' "end"
>        return (group (nest 3 (pk <> strip (empty : pss ++ [text ("_" ++ show n)])) <+> pke))
> ppType' (TVar n)
>   =  return $ text ("_" ++ show n)
> ppType' (TNatString n)
>   =  return $ text ("_ns" ++ show n)

> ppActual (ActExpr e)
>   =  return (text "<expression>") -- preliminary
> ppActual (ActType ts)
>   =  do
>        pts <- withPrec 0 $ mapM ppType' ts
>        return (tbrackets (intersperse (text ", ") pts))

> pFormal f = runReader (ppFormal f) defaultPState

> ppFormal (ForPat p)
>   =  ppPat p
> ppFormal (ForTypePat xs)
> --  =  tbrackets (intersperse (text ", ") (map ltext xs))
>   =  do
>        pxs <- mapM (\ x -> do
>                              px <- ppTyName x
>                              pk <- ppKind
>                              return (px <> text " :" <+> pk)) xs
>        return (tbrackets (intersperse (text ", ") pxs))

> tbrackets d = block 1 (string "<" <> d <> string ">")

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Pretty printing: kinds}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> ppKind
>   =  do
>        pc <- ppCol askind 
>        return (pc (text "Type"))

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Pretty printing: interfaces}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> pSpec :: Bool -> Bool -> Spec -> Doc
> pSpec c d s = runReader (ppSpec s) (defaultPState { pcolor = c, pdebug = d })

> ppSpec :: Spec -> P Doc
> ppSpec (SpVal x t)
>   =  do
>        pk <- keyword "val"
>        pt <- withPrec 0 $ ppType t
>        return (group (pk <> block 5 (ltext x <> text " :" <+> pt)))
> ppSpec (SpType x (TyFun fs t))
>   =  do
>        -- print type arguments on the lhs
>        let (fs1,fs2) = span (\ f -> case f of
>                                       ForTypePat _ -> True
>                                       _            -> False) fs
>        pk <- keyword "type"
>        px <- ppType (TyCon x)
>        pr <- ppCol astype
>        pfs <- mapM ppFormal fs1
>        pt <- withPrec 0 $ ppType (TyFun fs2 t)
>        return (group (pk <> block 5 (px <> (if null fs1 then empty else text " " <> pr (strip pfs))
>                                     <> text " =" <+> pt))) 
> ppSpec (SpType x t)
>   =  do
>        pk <- keyword "type"
>        px <- ppType (TyCon x)
>        pt <- withPrec 0 $ ppType t
>        return (group (pk <> block 5 (px <> text " =" <+> pt)))
> ppSpec (SpData x)
>   =  do
>        pk <- keyword "data"
>        px <- ppType (TyCon x)
>        return (group (pk <> px))
> ppSpec (SpCon x (TyForall _ t))
>   =  do
>        pk <- keyword "con"
>        px <- ppLCon x
>        pt <- withPrec 0 $ ppType t
>        return (group (pk <> block 5 (px <> text " :" <+> pt)))
> ppSpec (SpCon x t)
>   =  ppSpec (SpCon x (TyForall [] t))
> ppSpec (SpException (x, Nothing))
>   =  do
>        pk <- keyword "exception"
>        px <- ppLCon x
>        return (pk <> px)
> ppSpec (SpException (x, Just t))
>   =  do
>        pk <- keyword "exception"
>        px <- ppLCon x
>        pt <- withPrec 0 $ ppType t
>        return (group (pk <> block 5 (px <+> pt)))
> ppSpec (SpModule x s)
>   =  do
>        pk <- keyword "module"
>        ps <- ppSig s
>        return (pk <> block 5 (ltext x <> text " :" <+> ps))
> ppSpec (SpInterface x s)
>   =  do
>        pk <- keyword "interface"
>        ps <- ppSig s
>        return (pk <> block 5 (ltext x <> text " =" <+> ps))
> ppSpec x = return (pretty x)

> ppSig (Sig s)            =  keyword' "sig"  -- not showing the contents of the sig
> ppSig (SigFun [] s)      =  ppSig s
> ppSig (SigFun fs s)
>   =  do
>        pk <- keyword "fun"
>        pfs <- mapM ppSigFormal fs
>        ps <- ppSig s
>        return (group (pk <> strip pfs <> text " =>" <+> ps))
> ppSig (SigFunctor [] s)  =  ppSig s
> ppSig (SigFunctor fs s)
>   =  do
>        pk <- keyword "functor"
>        pfs <- mapM ppStrFormal fs
>        ps <- ppSig s
>        return (group (pk <> strip pfs <> text " =>" <+> ps))
> ppSig x                  =  return (pretty x)

> ppSigFormal xs
>   =  do
>        pxs <- mapM (\ x -> do
>                              px <- ppTyName x
>                              pk <- ppKind
>                              return (px <> text " :" <+> pk)) xs
>        return (tbrackets (intersperse (text ", ") pxs))

> ppStrFormal (StrForStrPat x s)
>   =  do
>        ps <- ppSig s
>        return (parens (ltext x <> text " : " <> ps))
> ppStrFormal (StrForTypePat xs)
>   =  do
>        pxs <- mapM (\ x -> do
>                              px <- ppTyName x
>                              pk <- ppKind
>                              return (px <> text " :" <+> pk)) xs
>        return (tbrackets (intersperse (text ", ") pxs))

